@@ -27,14 +27,29 @@ from __future__ import print_function
 from __future__ import unicode_literals
 
 import logging
-from os import extsep, path, readlink, curdir
+from os import environ, extsep, path, readlink
 from subprocess import CalledProcessError, Popen, PIPE
 import sys
-import tarfile
-from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED
 import re
 
-__version__ = "1.16.4"
+__version__ = "1.19.4"
+
+
+try:
+    # Python 3.3+
+    from shlex import quote
+except ImportError:
+    _find_unsafe = re.compile(r'[^a-zA-Z0-9_@%+=:,./-]').search
+
+    def quote(s):
+        """Return a shell-escaped version of the string *s*."""
+        if not s:
+            return "''"
+
+        if _find_unsafe(s) is None:
+            return s
+
+        return "'" + s.replace("'", "'\"'\"'") + "'"
 
 
 class GitArchiver(object):
@@ -48,9 +63,20 @@ class GitArchiver(object):
     >>> archiver = GitArchiver(main_repo_abspath='my/repo/path')
     >>> archiver.create('output.zip')
     """
+    TARFILE_FORMATS = {
+        'tar': 'w',
+        'tbz2': 'w:bz2',
+        'tgz': 'w:gz',
+        'txz': 'w:xz',
+        'bz2': 'w:bz2',
+        'gz': 'w:gz',
+        'xz': 'w:xz'
+    }
+    ZIPFILE_FORMATS = ('zip',)
+
     LOG = logging.getLogger('GitArchiver')
 
-    def __init__(self, prefix='', exclude=True, force_sub=False, extra=None, main_repo_abspath=None):
+    def __init__(self, prefix='', exclude=True, force_sub=False, extra=None, main_repo_abspath=None, git_version=None):
         """
         @param prefix: Prefix used to prepend all paths in the resulting archive.
             Extra file paths are only prefixed if they are not relative.
@@ -75,7 +101,17 @@ class GitArchiver(object):
             with abspath to top-level directory of the repository.
             If None, current cwd is used.
         @type main_repo_abspath: str
+
+        @param git_version: Version of Git that determines whether various workarounds are on.
+            If None, tries to resolve via Git's CLI.
+        @type git_version: tuple or None
         """
+        if git_version is None:
+            git_version = self.get_git_version()
+
+        if git_version is not None and git_version < (1, 6, 1):
+            raise ValueError("git of version 1.6.1 and higher is required")
+
         if extra is None:
             extra = []
 
@@ -94,8 +130,11 @@ class GitArchiver(object):
         self.extra = extra
         self.force_sub = force_sub
         self.main_repo_abspath = main_repo_abspath
+        self.git_version = git_version
 
-    def create(self, output_path, dry_run=False, output_format=None):
+        self._check_attr_gens = {}
+
+    def create(self, output_path, dry_run=False, output_format=None, compresslevel=None):
         """
         Create the archive at output_file_path.
 
@@ -118,8 +157,16 @@ class GitArchiver(object):
             self.LOG.debug("Output format is not explicitly set, determined format is {0}.".format(output_format))
 
         if not dry_run:
-            if output_format == 'zip':
-                archive = ZipFile(path.abspath(output_path), 'w')
+            if output_format in self.ZIPFILE_FORMATS:
+                from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED
+
+                if compresslevel is not None:
+                    if sys.version_info > (3, 7):
+                        archive = ZipFile(path.abspath(output_path), 'w', compresslevel=compresslevel)
+                    else:
+                        raise ValueError("Compression level for zip archives requires Python 3.7+")
+                else:
+                    archive = ZipFile(path.abspath(output_path), 'w')
 
                 def add_file(file_path, arcname):
                     if not path.islink(file_path):
@@ -129,25 +176,26 @@ class GitArchiver(object):
                         i.create_system = 3
                         i.external_attr = 0xA1ED0000
                         archive.writestr(i, readlink(file_path))
-            elif output_format in ['tar', 'bz2', 'gz', 'xz', 'tgz', 'txz']:
-                if output_format == 'tar':
-                    t_mode = 'w'
-                elif output_format == 'tgz':
-                    t_mode = 'w:gz'
-                elif output_format == 'txz':
-                    t_mode = 'w:xz'
-                else:
-                    t_mode = 'w:{0}'.format(output_format)
+            elif output_format in self.TARFILE_FORMATS:
+                import tarfile
 
-                archive = tarfile.open(path.abspath(output_path), t_mode)
+                mode = self.TARFILE_FORMATS[output_format]
+
+                if compresslevel is not None:
+                    try:
+                        archive = tarfile.open(path.abspath(output_path), mode, compresslevel=compresslevel)
+                    except TypeError:
+                        raise ValueError("{0} cannot be compressed".format(output_format))
+                else:
+                    archive = tarfile.open(path.abspath(output_path), mode)
 
                 def add_file(file_path, arcname):
                     archive.add(file_path, arcname)
             else:
-                raise RuntimeError("unknown format: {0}".format(output_format))
+                raise ValueError("unknown format: {0}".format(output_format))
 
             def archiver(file_path, arcname):
-                self.LOG.debug("Compressing {0} => {1}...".format(file_path, arcname))
+                self.LOG.debug("{0} => {1}".format(file_path, arcname))
                 add_file(file_path, arcname)
         else:
             archive = None
@@ -160,105 +208,22 @@ class GitArchiver(object):
         if archive is not None:
             archive.close()
 
-    def get_exclude_patterns(self, repo_abspath, repo_file_paths):
-        """
-        Returns exclude patterns for a given repo. It looks for .gitattributes files in repo_file_paths.
-
-        Resulting dictionary will contain exclude patterns per path (relative to the repo_abspath).
-        E.g. {('.', 'Catalyst', 'Editions', 'Base'): ['Foo*', '*Bar']}
-
-        @param repo_abspath: Absolute path to the git repository.
-        @type repo_abspath: str
-
-        @param repo_file_paths: List of paths relative to the repo_abspath that are under git control.
-        @type repo_file_paths:  list
-
-        @return: Dictionary representing exclude patterns.
-            Keys are tuples of strings. Values are lists of strings.
-            Returns None if self.exclude is not set.
-        @rtype: dict or None
-        """
-        if not self.exclude:
-            return None
-
-        def read_attributes(attributes_abspath):
-            patterns = []
-            if path.isfile(attributes_abspath):
-                attributes = open(attributes_abspath, 'r').readlines()
-                patterns = []
-                for line in attributes:
-                    tokens = line.strip().split()
-                    if "export-ignore" in tokens[1:]:
-                        patterns.append(tokens[0])
-            return patterns
-
-        exclude_patterns = {(): []}
-
-        # There may be no gitattributes.
-        try:
-            global_attributes_abspath = self.run_git_shell("git config --get core.attributesfile", repo_abspath).rstrip()
-            exclude_patterns[()] = read_attributes(global_attributes_abspath)
-        except:
-            # And it's valid to not have them.
-            pass
-
-        for attributes_abspath in [path.join(repo_abspath, f) for f in repo_file_paths if f.endswith(".gitattributes")]:
-            # Each .gitattributes affects only files within its directory.
-            key = tuple(self.get_path_components(repo_abspath, path.dirname(attributes_abspath)))
-            exclude_patterns[key] = read_attributes(attributes_abspath)
-
-        local_attributes_abspath = path.join(repo_abspath, ".git", "info", "attributes")
-        key = tuple(self.get_path_components(repo_abspath, repo_abspath))
-
-        if key in exclude_patterns:
-            exclude_patterns[key].extend(read_attributes(local_attributes_abspath))
-        else:
-            exclude_patterns[key] = read_attributes(local_attributes_abspath)
-
-        return exclude_patterns
-
-    def is_file_excluded(self, repo_abspath, repo_file_path, exclude_patterns):
+    def is_file_excluded(self, repo_abspath, repo_file_path):
         """
         Checks whether file at a given path is excluded.
 
         @param repo_abspath: Absolute path to the git repository.
         @type repo_abspath: str
 
-        @param repo_file_path: Path to a file within repo_abspath.
+        @param repo_file_path: Path to a file relative to repo_abspath.
         @type repo_file_path: str
 
-        @param exclude_patterns: Exclude patterns with format specified for get_exclude_patterns.
-        @type exclude_patterns: dict
-
         @return: True if file should be excluded. Otherwise False.
         @rtype: bool
         """
-        if exclude_patterns is None or not len(exclude_patterns):
-            return False
-
-        from fnmatch import fnmatch
-
-        file_name = path.basename(repo_file_path)
-        components = self.get_path_components(repo_abspath, path.join(repo_abspath, path.dirname(repo_file_path)))
-
-        is_excluded = False
-        # We should check all patterns specified in intermediate directories to the given file.
-        # At the end we should also check for the global patterns (key '()' or empty tuple).
-        while not is_excluded:
-            key = tuple(components)
-            if key in exclude_patterns:
-                patterns = exclude_patterns[key]
-                for p in patterns:
-                    if fnmatch(file_name, p) or fnmatch(repo_file_path, p):
-                        self.LOG.debug("Exclude pattern matched {0}: {1}".format(p, repo_file_path))
-                        is_excluded = True
-
-            if not len(components):
-                break
-
-            components.pop()
-
-        return is_excluded
+        next(self._check_attr_gens[repo_abspath])
+        attrs = self._check_attr_gens[repo_abspath].send(repo_file_path)
+        return attrs['export-ignore'] == 'set'
 
     def archive_all_files(self, archiver):
         """
@@ -290,98 +255,177 @@ class GitArchiver(object):
         @rtype: Iterable
         """
         repo_abspath = path.join(self.main_repo_abspath, repo_path)
-        repo_file_paths = self.run_git_shell(
-            "git ls-files --cached --full-name --no-empty-directory",
-            repo_abspath
-        ).splitlines()
-        exclude_patterns = self.get_exclude_patterns(repo_abspath, repo_file_paths)
-
-        for repo_file_path in repo_file_paths:
-            # Git puts path in quotes if file path has unicode characters.
-            repo_file_path = repo_file_path.strip('"')  # file path relative to current repo
-            repo_file_abspath = path.join(repo_abspath, repo_file_path)  # absolute file path
-            main_repo_file_path = path.join(repo_path, repo_file_path)  # file path relative to the main repo
-
-            # Only list symlinks and files.
-            if not path.islink(repo_file_abspath) and path.isdir(repo_file_abspath):
-                continue
-
-            if self.is_file_excluded(repo_abspath, repo_file_path, exclude_patterns):
-                continue
-
-            yield main_repo_file_path
-
-        if self.force_sub:
-            self.run_git_shell("git submodule init", repo_abspath)
-            self.run_git_shell("git submodule update", repo_abspath)
+        assert repo_abspath not in self._check_attr_gens
+        self._check_attr_gens[repo_abspath] = self.check_attr(repo_abspath, ['export-ignore'])
 
         try:
-            repo_gitmodules_abspath = path.join(repo_abspath, ".gitmodules")
+            repo_file_paths = self.run_git_shell(
+                'git ls-files -z --cached --full-name --no-empty-directory',
+                repo_abspath
+            ).split('\0')[:-1]
 
-            with open(repo_gitmodules_abspath) as f:
-                lines = f.readlines()
+            for repo_file_path in repo_file_paths:
+                repo_file_abspath = path.join(repo_abspath, repo_file_path)  # absolute file path
+                main_repo_file_path = path.join(repo_path, repo_file_path)  # relative to main_repo_abspath
 
-            for l in lines:
-                m = re.match("^\s*path\s*=\s*(.*)\s*$", l)
+                # Only list symlinks and files.
+                if not path.islink(repo_file_abspath) and path.isdir(repo_file_abspath):
+                    continue
 
-                if m:
-                    submodule_path = m.group(1)
-                    submodule_path = path.join(repo_path, submodule_path)
+                if self.is_file_excluded(repo_abspath, repo_file_path):
+                    continue
 
-                    for submodule_file_path in self.walk_git_files(submodule_path):
-                        yield submodule_file_path
-        except IOError:
-            pass
+                yield main_repo_file_path
 
-    @staticmethod
-    def get_path_components(repo_abspath, abspath):
+            if self.force_sub:
+                self.run_git_shell('git submodule init', repo_abspath)
+                self.run_git_shell('git submodule update', repo_abspath)
+
+            try:
+                repo_gitmodules_abspath = path.join(repo_abspath, ".gitmodules")
+
+                with open(repo_gitmodules_abspath) as f:
+                    lines = f.readlines()
+
+                for l in lines:
+                    m = re.match("^\\s*path\\s*=\\s*(.*)\\s*$", l)
+
+                    if m:
+                        repo_submodule_path = m.group(1)  # relative to repo_path
+                        main_repo_submodule_path = path.join(repo_path, repo_submodule_path)  # relative to main_repo_abspath
+
+                        if self.is_file_excluded(repo_abspath, repo_submodule_path):
+                            continue
+
+                        for main_repo_submodule_file_path in self.walk_git_files(main_repo_submodule_path):
+                            repo_submodule_file_path = path.relpath(main_repo_submodule_file_path, repo_path)  # relative to repo_path
+                            if self.is_file_excluded(repo_abspath, repo_submodule_file_path):
+                                continue
+
+                            yield main_repo_submodule_file_path
+            except IOError:
+                pass
+        finally:
+            self._check_attr_gens[repo_abspath].close()
+            del self._check_attr_gens[repo_abspath]
+
+    def check_attr(self, repo_abspath, attrs):
         """
-        Split given abspath into components relative to repo_abspath.
-        These components are primarily used as unique keys of files and folders within a repository.
+        Generator that returns attributes for given paths relative to repo_abspath.
 
-        E.g. if repo_abspath is '/Documents/Hobby/ParaView/' and abspath is
-        '/Documents/Hobby/ParaView/Catalyst/Editions/Base/', function will return:
-        ['.', 'Catalyst', 'Editions', 'Base']
+        >>> g = GitArchiver.check_attr('repo_path', ['export-ignore'])
+        >>> next(g)
+        >>> attrs = g.send('relative_path')
+        >>> print(attrs['export-ignore'])
 
-        First element is always os.curdir (concrete symbol depends on OS).
-
-        @param repo_abspath: Absolute path to the git repository. Normalized via os.path.normpath.
+        @param repo_abspath: Absolute path to a git repository.
         @type repo_abspath: str
 
-        @param abspath: Absolute path to a file within repo_abspath. Normalized via os.path.normpath.
-        @type abspath: str
+        @param attrs: Attributes to check.
+        @type attrs: [str]
 
-        @return: List of path components.
-        @rtype: list
+        @rtype: generator
         """
-        repo_abspath = path.normpath(repo_abspath)
-        abspath = path.normpath(abspath)
+        def make_process():
+            env = dict(environ, GIT_FLUSH='1')
+            cmd = 'git check-attr --stdin -z {0}'.format(' '.join(attrs))
+            return Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE, cwd=repo_abspath, env=env)
 
-        if not path.isabs(repo_abspath):
-            raise ValueError("repo_abspath MUST be absolute path.")
+        def read_attrs(process, repo_file_path):
+            process.stdin.write(repo_file_path.encode('utf-8') + b'\0')
+            process.stdin.flush()
 
-        if not path.isabs(abspath):
-            raise ValueError("abspath MUST be absoulte path.")
+            # For every attribute check-attr will output: <path> NUL <attribute> NUL <info> NUL
+            path, attr, info = b'', b'', b''
+            nuls_count = 0
+            nuls_expected = 3 * len(attrs)
 
-        if not path.commonprefix([repo_abspath, abspath]):
-            raise ValueError(
-                "abspath (\"{0}\") MUST have common prefix with repo_abspath (\"{1}\")"
-                .format(abspath, repo_abspath)
-            )
+            while nuls_count != nuls_expected:
+                b = process.stdout.read(1)
 
-        components = []
+                if b == b'' and process.poll() is not None:
+                    raise RuntimeError("check-attr exited prematurely")
+                elif b == b'\0':
+                    nuls_count += 1
 
-        while not abspath == repo_abspath:
-            abspath, tail = path.split(abspath)
+                    if nuls_count % 3 == 0:
+                        yield map(self.decode_git_output, (path, attr, info))
+                        path, attr, info = b'', b'', b''
+                elif nuls_count % 3 == 0:
+                    path += b
+                elif nuls_count % 3 == 1:
+                    attr += b
+                elif nuls_count % 3 == 2:
+                    info += b
 
-            if tail:
-                components.insert(0, tail)
+        def read_attrs_old(process, repo_file_path):
+            """
+            Compatibility with versions 1.8.5 and below that do not recognize -z for output.
+            """
+            process.stdin.write(repo_file_path.encode('utf-8') + b'\0')
+            process.stdin.flush()
 
-        components.insert(0, curdir)
-        return components
+            # For every attribute check-attr will output: <path>: <attribute>: <info>\n
+            # where <path> is c-quoted
 
-    @staticmethod
-    def run_git_shell(cmd, cwd=None):
+            path, attr, info = b'', b'', b''
+            lines_count = 0
+            lines_expected = len(attrs)
+
+            while lines_count != lines_expected:
+                line = process.stdout.readline()
+
+                info_start = line.rfind(b': ')
+                if info_start == -1:
+                    raise RuntimeError("unexpected output of check-attr: {0}".format(line))
+
+                attr_start = line.rfind(b': ', 0, info_start)
+                if attr_start == -1:
+                    raise RuntimeError("unexpected output of check-attr: {0}".format(line))
+
+                info = line[info_start + 2:len(line) - 1]  # trim leading ": " and trailing \n
+                attr = line[attr_start + 2:info_start]  # trim leading ": "
+                path = line[:attr_start]
+
+                yield map(self.decode_git_output, (path, attr, info))
+                lines_count += 1
+
+        if not attrs:
+            return
+
+        process = make_process()
+
+        try:
+            while True:
+                repo_file_path = yield
+                repo_file_attrs = {}
+
+                if self.git_version is None or self.git_version > (1, 8, 5):
+                    reader = read_attrs
+                else:
+                    reader = read_attrs_old
+
+                for path, attr, value in reader(process, repo_file_path):
+                    repo_file_attrs[attr] = value
+
+                yield repo_file_attrs
+        finally:
+            process.stdin.close()
+            process.wait()
+
+    @classmethod
+    def decode_git_output(cls, output):
+        """
+        Decode Git's binary output handeling the way it escapes unicode characters.
+
+        @type output: bytes
+
+        @rtype: str
+        """
+        return output.decode('unicode_escape').encode('raw_unicode_escape').decode('utf-8')
+
+    @classmethod
+    def run_git_shell(cls, cmd, cwd=None):
         """
         Runs git shell command, reads output and decodes it into unicode string.
 
@@ -398,7 +442,7 @@ class GitArchiver(object):
         """
         p = Popen(cmd, shell=True, stdout=PIPE, cwd=cwd)
         output, _ = p.communicate()
-        output = output.decode('unicode_escape').encode('raw_unicode_escape').decode('utf-8')
+        output = cls.decode_git_output(output)
 
         if p.returncode:
             if sys.version_info > (2, 6):
@@ -408,13 +452,40 @@ class GitArchiver(object):
 
         return output
 
+    @classmethod
+    def get_git_version(cls):
+        """
+        Return version of git current shell points to.
+
+        If version cannot be parsed None is returned.
+
+        @rtype: tuple or None
+        """
+        try:
+            output = cls.run_git_shell('git version')
+        except CalledProcessError:
+            cls.LOG.warning("Unable to get Git version.")
+            return None
+
+        try:
+            version = output.split()[2]
+        except IndexError:
+            cls.LOG.warning("Unable to parse Git version \"%s\".", output)
+            return None
+
+        try:
+            return tuple(int(v) for v in version.split('.'))
+        except ValueError:
+            cls.LOG.warning("Unable to parse Git version \"%s\".", version)
+            return None
+
 
 def main():
-    from optparse import OptionParser
+    from optparse import OptionParser, SUPPRESS_HELP
 
     parser = OptionParser(
         usage="usage: %prog [-v] [--prefix PREFIX] [--no-exclude] [--force-submodules]"
-              " [--extra EXTRA1 [EXTRA2]] [--dry-run] OUTPUT_FILE",
+              " [--extra EXTRA1 ...] [--dry-run] [-0 | ... | -9] OUTPUT_FILE",
         version="%prog {0}".format(__version__)
     )
 
@@ -453,6 +524,13 @@ def main():
                       dest='dry_run',
                       help="don't actually archive anything, just show what would be done")
 
+    for i in range(10):
+        parser.add_option('-{0}'.format(i),
+                          action='store_const',
+                          const=i,
+                          dest='compresslevel',
+                          help=SUPPRESS_HELP)
+
     options, args = parser.parse_args()
 
     if len(args) != 1:
@@ -467,11 +545,9 @@ def main():
     if options.prefix is not None:
         options.prefix = path.join(options.prefix, '')
     else:
-        import re
-
         output_name = path.basename(output_file_path)
         output_name = re.sub(
-            '(\.zip|\.tar|\.tgz|\.txz|\.gz|\.bz2|\.xz|\.tar\.gz|\.tar\.bz2|\.tar\.xz)$',
+            '(\\.zip|\\.tar|\\.tbz2|\\.tgz|\\.txz|\\.bz2|\\.gz|\\.xz|\\.tar\\.bz2|\\.tar\\.gz|\\.tar\\.xz)$',
             '',
             output_name
         ) or "Archive"
@@ -486,7 +562,7 @@ def main():
                                options.exclude,
                                options.force_sub,
                                options.extra)
-        archiver.create(output_file_path, options.dry_run)
+        archiver.create(output_file_path, options.dry_run, compresslevel=options.compresslevel)
     except Exception as e:
         parser.exit(2, "{0}\n".format(e))
 
